from _typeshed import Incomplete


class TempDoctest:
    modname: Incomplete
    docstr: Incomplete
    temp: Incomplete
    dpath: Incomplete
    modpath: Incomplete

    def __init__(self, docstr, modname: Incomplete | None = ...) -> None:
        ...


class TempModule:
    modname: Incomplete
    module_text: Incomplete
    temp: Incomplete
    dpath: Incomplete
    modpath: Incomplete

    def __init__(self, module_text, modname: Incomplete | None = ...) -> None:
        ...
