from xdoctest.docstr import docscrape_google

from xdoctest.docstr.docscrape_google import (parse_google_argblock,
                                              parse_google_args,
                                              parse_google_retblock,
                                              parse_google_returns,
                                              split_google_docblocks,)

__all__ = ['docscrape_google', 'parse_google_argblock', 'parse_google_args',
           'parse_google_retblock', 'parse_google_returns',
           'split_google_docblocks']
